﻿namespace Microsoft.VisualStudio.TextTemplating.CodeDom
{
    using System;
    using System.CodeDom;
    using System.Collections.Generic;
    using System.Reflection;
    using System.Runtime.CompilerServices;

    internal static class CodeDomHelpers
    {
        internal static readonly CodePrimitiveExpression nullEx = new CodePrimitiveExpression(null);

        internal static CodeBinaryOperatorExpression Add(this CodeExpression lhs, CodeExpression rhs)
        {
            return new CodeBinaryOperatorExpression(lhs, CodeBinaryOperatorType.Add, rhs);
        }

        internal static void AddSummaryComment(this CodeTypeMember member, string comment)
        {
            member.Comments.Add(new CodeCommentStatement(new CodeComment("<summary>", true)));
            member.Comments.Add(new CodeCommentStatement(new CodeComment(comment, true)));
            member.Comments.Add(new CodeCommentStatement(new CodeComment("</summary>", true)));
        }

        internal static CodeBinaryOperatorExpression And(this CodeExpression lhs, CodeExpression rhs)
        {
            return new CodeBinaryOperatorExpression(lhs, CodeBinaryOperatorType.BooleanAnd, rhs);
        }

        internal static CodeAssignStatement Assign(this CodeExpression lhs, CodeExpression rhs)
        {
            return new CodeAssignStatement(lhs, rhs);
        }

        internal static CodeMethodInvokeExpression Call(string methodName, params CodeExpression[] parameters)
        {
            return new CodeMethodInvokeExpression(new CodeThisReferenceExpression(), methodName, parameters);
        }

        internal static CodeMethodInvokeExpression Call(this CodeExpression callSite, string methodName, params CodeExpression[] parameters)
        {
            return new CodeMethodInvokeExpression(new CodeMethodReferenceExpression(callSite, methodName), parameters);
        }

        internal static CodeMethodInvokeExpression Call(string callSite, string methodName, params CodeExpression[] parameters)
        {
            return new CodeVariableReferenceExpression(callSite).Call(methodName, parameters);
        }

        internal static CodeMethodInvokeExpression Call(Type type, string methodName, params CodeExpression[] parameters)
        {
            return new CodeMethodInvokeExpression(new CodeMethodReferenceExpression(type.Expr(), methodName), parameters);
        }

        internal static CodeExpressionStatement CallS(string methodName, params CodeExpression[] parameters)
        {
            return new CodeExpressionStatement(Call(methodName, parameters));
        }

        internal static CodeExpressionStatement CallS(this CodeExpression callSite, string methodName, params CodeExpression[] parameters)
        {
            return new CodeExpressionStatement(callSite.Call(methodName, parameters));
        }

        internal static CodeExpressionStatement CallS(string callSite, string methodName, params CodeExpression[] parameters)
        {
            return new CodeExpressionStatement(Call(callSite, methodName, parameters));
        }

        internal static CodeExpressionStatement CallS(Type type, string methodName, params CodeExpression[] parameters)
        {
            return new CodeExpressionStatement(Call(type, methodName, parameters));
        }

        internal static CodeCastExpression Cast(this CodeTypeReference targetType, CodeExpression value)
        {
            return new CodeCastExpression(targetType, value);
        }

        internal static CodeCastExpression Cast(this Type targetType, CodeExpression value)
        {
            return new CodeCastExpression(targetType.Ref(), value);
        }

        internal static CodeConditionStatement CheckNullParameter(string name)
        {
            return new CodeConditionStatement(new CodeBinaryOperatorExpression(new CodeVariableReferenceExpression(name), CodeBinaryOperatorType.IdentityEquality, nullEx), new CodeStatement[] { new CodeThrowExceptionStatement(typeof(ArgumentNullException).New(new CodeExpression[] { name.Prim() })) });
        }

        internal static CodeMemberMethod CreateMethod(Type returnType, string name, string summaryComment, MemberAttributes attributes, params CodeObject[] statements)
        {
            CodeMemberMethod member = new CodeMemberMethod {
                ReturnType = (returnType == null) ? null : returnType.Ref(),
                Name = name
            };
            member.AddSummaryComment(summaryComment);
            member.Attributes = attributes;
            foreach (CodeObject obj2 in statements)
            {
                CodeStatement statement = obj2 as CodeStatement;
                CodeExpression expression = obj2 as CodeExpression;
                if (statement != null)
                {
                    member.Statements.Add(statement);
                }
                if (expression != null)
                {
                    member.Statements.Add(expression);
                }
            }
            return member;
        }

        internal static CodeVariableDeclarationStatement CreatePropertyInfoDeclaration(string name, CodeExpression target, string propertyName)
        {
            return new CodeVariableDeclarationStatement(typeof(PropertyInfo), name, target.Call("GetType", new CodeExpression[0]).Call("GetProperty", new CodeExpression[] { propertyName.Prim() }));
        }

        internal static CodeVariableDeclarationStatement CreatePropertyInfoValueDeclaration(string name, CodeExpression propertyInfo, CodeExpression instance, params CodeExpression[] indexParameters)
        {
            List<CodeExpression> list = new List<CodeExpression> {
                instance
            };
            if (indexParameters.Length == 0)
            {
                list.Add(nullEx);
            }
            else
            {
                list.Add(new CodeArrayCreateExpression(typeof(object), indexParameters));
            }
            return new CodeVariableDeclarationStatement(typeof(object), name, propertyInfo.Call("GetValue", list.ToArray()));
        }

        internal static CodeVariableDeclarationStatement Decl(this Type type, string name, CodeExpression initializer)
        {
            return new CodeVariableDeclarationStatement(type.Ref(), name, initializer);
        }

        internal static CodeTypeReferenceExpression Expr(this Type type)
        {
            return new CodeTypeReferenceExpression(type) { Type = { Options = CodeTypeReferenceOptions.GlobalReference } };
        }

        internal static CodeBinaryOperatorExpression Gt(this CodeExpression lhs, CodeExpression rhs)
        {
            return new CodeBinaryOperatorExpression(lhs, CodeBinaryOperatorType.GreaterThan, rhs);
        }

        internal static CodeBinaryOperatorExpression IEquals(this CodeExpression lhs, CodeExpression rhs)
        {
            return new CodeBinaryOperatorExpression(lhs, CodeBinaryOperatorType.IdentityEquality, rhs);
        }

        internal static CodeConditionStatement IfNotNull(this CodeVariableReferenceExpression variable, params CodeStatement[] trueStatements)
        {
            return IfVariableNotNull(variable, trueStatements);
        }

        internal static CodeConditionStatement IfVariableNotNull(CodeVariableReferenceExpression variable, params CodeStatement[] trueStatements)
        {
            return new CodeConditionStatement(variable.INEquals(nullEx), trueStatements);
        }

        internal static CodeConditionStatement IfVariableNullThenInstantiateType(CodeFieldReferenceExpression variableRef, Type type)
        {
            return new CodeConditionStatement(new CodeBinaryOperatorExpression(variableRef, CodeBinaryOperatorType.IdentityEquality, nullEx), new CodeStatement[] { new CodeAssignStatement(variableRef, type.New()) });
        }

        internal static CodeIndexerExpression Index(this CodeExpression expression, params CodeExpression[] indices)
        {
            return new CodeIndexerExpression(expression, indices);
        }

        internal static CodeBinaryOperatorExpression INEquals(this CodeExpression lhs, CodeExpression rhs)
        {
            return new CodeBinaryOperatorExpression(lhs, CodeBinaryOperatorType.IdentityInequality, rhs);
        }

        internal static CodeBinaryOperatorExpression IsNull(this CodeExpression operand)
        {
            return operand.IEquals(nullEx);
        }

        internal static CodeBinaryOperatorExpression Lt(this CodeExpression lhs, CodeExpression rhs)
        {
            return new CodeBinaryOperatorExpression(lhs, CodeBinaryOperatorType.LessThan, rhs);
        }

        internal static CodeObjectCreateExpression New(this Type type)
        {
            return new CodeObjectCreateExpression(type.Ref(), new CodeExpression[0]);
        }

        internal static CodeObjectCreateExpression New(this Type type, params CodeExpression[] parameters)
        {
            return new CodeObjectCreateExpression(type.Ref(), parameters);
        }

        internal static CodeBinaryOperatorExpression NotNull(this CodeExpression operand)
        {
            return operand.INEquals(nullEx);
        }

        internal static CodeBinaryOperatorExpression Or(this CodeExpression lhs, CodeExpression rhs)
        {
            return new CodeBinaryOperatorExpression(lhs, CodeBinaryOperatorType.BooleanAnd, rhs);
        }

        internal static CodePrimitiveExpression Prim(this object primitive)
        {
            return new CodePrimitiveExpression(primitive);
        }

        internal static CodePropertyReferenceExpression Prop(this CodeExpression site, string property)
        {
            return new CodePropertyReferenceExpression(site, property);
        }

        internal static CodeFieldReferenceExpression Ref(this CodeMemberField field)
        {
            return new CodeFieldReferenceExpression(new CodeThisReferenceExpression(), field.Name);
        }

        internal static CodePropertyReferenceExpression Ref(this CodeMemberProperty property)
        {
            return new CodePropertyReferenceExpression(new CodeThisReferenceExpression(), property.Name);
        }

        internal static CodeVariableReferenceExpression Ref(this CodeParameterDeclarationExpression parameter)
        {
            return new CodeVariableReferenceExpression(parameter.Name);
        }

        internal static CodeVariableReferenceExpression Ref(this CodeVariableDeclarationStatement declaration)
        {
            return new CodeVariableReferenceExpression(declaration.Name);
        }

        internal static CodeTypeReference Ref(this Type type)
        {
            return new CodeTypeReference(type, CodeTypeReferenceOptions.GlobalReference);
        }

        internal static CodeBinaryOperatorExpression Subtract(this CodeExpression lhs, CodeExpression rhs)
        {
            return new CodeBinaryOperatorExpression(lhs, CodeBinaryOperatorType.Subtract, rhs);
        }

        internal static CodeBinaryOperatorExpression UnaryNot(this CodeExpression operand)
        {
            return new CodeBinaryOperatorExpression(operand, CodeBinaryOperatorType.ValueEquality, false.Prim());
        }

        internal static CodeBinaryOperatorExpression VEquals(this CodeExpression lhs, CodeExpression rhs)
        {
            return new CodeBinaryOperatorExpression(lhs, CodeBinaryOperatorType.ValueEquality, rhs);
        }

        internal static CodeBinaryOperatorExpression VNEquals(this CodeExpression lhs, CodeExpression rhs)
        {
            return new CodeBinaryOperatorExpression(new CodeBinaryOperatorExpression(lhs, CodeBinaryOperatorType.ValueEquality, rhs), CodeBinaryOperatorType.ValueEquality, false.Prim());
        }
    }
}

